@@ -2,7 +2,7 @@
 
 /*  convert.c -- functions for converting between VCF/BCF and related formats.
 
-    Copyright (C) 2013-2021 Genome Research Ltd.
+    Copyright (C) 2013-2024 Genome Research Ltd.
 
     Author: Petr Danecek <pd3@sanger.ac.uk>
 
@@ -33,6 +33,7 @@ THE SOFTWARE.  */
 #include <errno.h>
 #include <sys/stat.h>
 #include <sys/types.h>
+#include <stdint.h>
 #define __STDC_FORMAT_MACROS
 #include <inttypes.h>
 #include <math.h>
@@ -40,6 +41,8 @@ THE SOFTWARE.  */
 #include <htslib/synced_bcf_reader.h>
 #include <htslib/vcfutils.h>
 #include <htslib/kfunc.h>
+#include <htslib/khash_str2int.h>
+#include <htslib/hts_endian.h>
 #include "bcftools.h"
 #include "variantkey.h"
 #include "convert.h"
@@ -103,7 +106,14 @@ struct _convert_t
     void *dat;
     int ndat;
     char *undef_info_tag;
+    void *used_tags_hash;
+    char **used_tags_list;
+    char *print_filtered;
+    int nused_tags;
     int allow_undef_tags;
+    int force_newline;
+    int header_samples;
+    int no_hdr_indices;
     uint8_t **subset_samples;
 };
 
@@ -169,23 +179,23 @@ static void process_filter(convert_t *convert, bcf1_t *line, fmt_t *fmt, int isa
     }
     else kputc('.', str);
 }
-static inline int32_t bcf_array_ivalue(void *bcf_array, int type, int idx)
+static inline int32_t bcf_array_ivalue(uint8_t *bcf_array, int type, int idx)
 {
     if ( type==BCF_BT_INT8 )
     {
-        int8_t val = ((int8_t*)bcf_array)[idx];
+        int8_t val = le_to_i8(&bcf_array[idx * sizeof(val)]);
         if ( val==bcf_int8_missing ) return bcf_int32_missing;
         if ( val==bcf_int8_vector_end ) return bcf_int32_vector_end;
         return val;
     }
     if ( type==BCF_BT_INT16 )
     {
-        int16_t val = ((int16_t*)bcf_array)[idx];
+        int16_t val = le_to_i16(&bcf_array[idx * sizeof(val)]);
         if ( val==bcf_int16_missing ) return bcf_int32_missing;
         if ( val==bcf_int16_vector_end ) return bcf_int32_vector_end;
         return val;
     }
-    return ((int32_t*)bcf_array)[idx];
+    return le_to_i32(&bcf_array[idx * sizeof(int32_t)]);
 }
 static inline void _copy_field(char *src, uint32_t len, int idx, kstring_t *str)
 {
@@ -283,17 +293,17 @@ static void process_info(convert_t *convert, bcf1_t *line, fmt_t *fmt, int isamp
             kputc('.', str);
             return;
         }
-        #define BRANCH(type_t, is_missing, is_vector_end, kprint) { \
-            type_t val = ((type_t *) info->vptr)[fmt->subscript]; \
+        #define BRANCH(type_t, convert, is_missing, is_vector_end, kprint) { \
+            type_t val = convert(&info->vptr[fmt->subscript * sizeof(type_t)]); \
             if ( is_missing || is_vector_end ) kputc('.',str); \
             else kprint; \
         }
         switch (info->type)
         {
-            case BCF_BT_INT8:  BRANCH(int8_t,  val==bcf_int8_missing,  val==bcf_int8_vector_end,  kputw(val, str)); break;
-            case BCF_BT_INT16: BRANCH(int16_t, val==bcf_int16_missing, val==bcf_int16_vector_end, kputw(val, str)); break;
-            case BCF_BT_INT32: BRANCH(int32_t, val==bcf_int32_missing, val==bcf_int32_vector_end, kputw(val, str)); break;
-            case BCF_BT_FLOAT: BRANCH(float,   bcf_float_is_missing(val), bcf_float_is_vector_end(val), kputd(val, str)); break;
+            case BCF_BT_INT8:  BRANCH(int8_t,  le_to_i8,  val==bcf_int8_missing,  val==bcf_int8_vector_end,  kputw(val, str)); break;
+            case BCF_BT_INT16: BRANCH(int16_t, le_to_i16, val==bcf_int16_missing, val==bcf_int16_vector_end, kputw(val, str)); break;
+            case BCF_BT_INT32: BRANCH(int32_t, le_to_i32, val==bcf_int32_missing, val==bcf_int32_vector_end, kputw(val, str)); break;
+            case BCF_BT_FLOAT: BRANCH(float,   le_to_float, bcf_float_is_missing(val), bcf_float_is_vector_end(val), kputd(val, str)); break;
             case BCF_BT_CHAR:  _copy_field((char*)info->vptr, info->vptr_len, fmt->subscript, str); break;
             default: fprintf(bcftools_stderr,"todo: type %d\n", info->type); bcftools_exit(1); break;
         }
@@ -381,11 +391,12 @@ static void process_format(convert_t *convert, bcf1_t *line, fmt_t *fmt, int isa
         }
         if ( fmt->fmt->type == BCF_BT_FLOAT )
         {
-            float *ptr = (float*)(fmt->fmt->p + isample*fmt->fmt->size);
-            if ( bcf_float_is_missing(ptr[fmt->subscript]) || bcf_float_is_vector_end(ptr[fmt->subscript]) )
+            uint8_t *ptr = fmt->fmt->p + isample*fmt->fmt->size;
+            float val = le_to_float(&ptr[fmt->subscript * sizeof(float)]);
+            if ( bcf_float_is_missing(val) || bcf_float_is_vector_end(val) )
                 kputc('.', str);
             else
-                kputd(ptr[fmt->subscript], str);
+                kputd(val, str);
         }
         else if ( fmt->fmt->type != BCF_BT_CHAR )
         {
@@ -498,14 +509,14 @@ static void process_tbcsq(convert_t *convert, bcf1_t *line, fmt_t *fmt, int isam
 
     int mask = fmt->subscript==0 ? 3 : 1;   // merge both haplotypes if subscript==0
 
-    #define BRANCH(type_t, nbits) { \
-        type_t *x = (type_t*)(fmt->fmt->p + isample*fmt->fmt->size); \
+    #define BRANCH(type_t, convert, nbits) { \
+        uint8_t *x = fmt->fmt->p + isample*fmt->fmt->size; \
         int i,j; \
         if ( fmt->subscript<=0 || fmt->subscript==1 ) \
         { \
             for (j=0; j < fmt->fmt->n; j++) \
             { \
-                type_t val = x[j]; \
+                type_t val = convert(&x[j * sizeof(type_t)]); \
                 if ( !val ) continue; \
                 for (i=0; i<nbits; i+=2) \
                     if ( val & (mask<<i) ) { kputs(csq->str[(j*30+i)/2], &csq->hap1); kputc_(',', &csq->hap1); } \
@@ -515,7 +526,7 @@ static void process_tbcsq(convert_t *convert, bcf1_t *line, fmt_t *fmt, int isam
         { \
             for (j=0; j < fmt->fmt->n; j++) \
             { \
-                type_t val = x[j]; \
+                type_t val = convert(&x[j * sizeof(type_t)]); \
                 if ( !val ) continue; \
                 for (i=1; i<nbits; i+=2) \
                     if ( val & (1<<i) ) { kputs(csq->str[(j*30+i)/2], &csq->hap2); kputc_(',', &csq->hap2); } \
@@ -524,9 +535,9 @@ static void process_tbcsq(convert_t *convert, bcf1_t *line, fmt_t *fmt, int isam
     }
     switch (fmt->fmt->type)
     {
-        case BCF_BT_INT8:  BRANCH(uint8_t, 8); break;
-        case BCF_BT_INT16: BRANCH(uint16_t,16); break;
-        case BCF_BT_INT32: BRANCH(uint32_t,30); break;  // 2 bytes unused to account for the reserved BCF values
+        case BCF_BT_INT8:  BRANCH(uint8_t,  le_to_u8,   8); break;
+        case BCF_BT_INT16: BRANCH(uint16_t, le_to_u16, 16); break;
+        case BCF_BT_INT32: BRANCH(uint32_t, le_to_u32, 30); break;  // 2 bits unused to account for the reserved BCF values
         default: error("Unexpected type: %d\n", fmt->fmt->type); bcftools_exit(1); break;
     }
     #undef BRANCH
@@ -646,6 +657,7 @@ static void process_type(convert_t *convert, bcf1_t *line, fmt_t *fmt, int isamp
 static void process_line(convert_t *convert, bcf1_t *line, fmt_t *fmt, int isample, kstring_t *str)
 {
     vcf_format1(convert->header, line, str);
+    if ( str->s[str->l-1]=='\n' ) str->l--;
 }
 static void process_chrom_pos_id(convert_t *convert, bcf1_t *line, fmt_t *fmt, int isample, kstring_t *str)
 {
@@ -783,14 +795,13 @@ static void process_gp_to_prob3(convert_t *convert, bcf1_t *line, fmt_t *fmt, in
     n /= convert->nsamples;
     for (i=0; i<convert->nsamples; i++)
     {
-        float sum = 0, *ptr = (float*)convert->dat + i*n;
+        float *ptr = (float*)convert->dat + i*n;
         int j;
         for (j=0; j<n; j++)
         {
             if ( bcf_float_is_vector_end(ptr[j]) ) break;
             if ( bcf_float_is_missing(ptr[j]) ) { ptr[j]=0; continue; }
             if ( ptr[j]<0 || ptr[j]>1 ) error("[%s:%"PRId64":%f] GP value outside range [0,1]; bcftools convert expects the VCF4.3+ spec for the GP field encoding genotype posterior probabilities", bcf_seqname(convert->header,line),(int64_t) line->pos+1,ptr[j]);
-            sum+=ptr[j];
         }
         if ( j==line->n_allele )
             ksprintf(str," %f %f %f",ptr[0],0.,ptr[1]); // haploid
@@ -954,12 +965,12 @@ static void process_gt_to_hap(convert_t *convert, bcf1_t *line, fmt_t *fmt, int
         }
         else if ( bcf_gt_is_missing(ptr[0]) )
         {
-            if ( ptr[1]==bcf_int8_vector_end ) 
+            if ( ptr[1]==bcf_int8_vector_end )
             {
                 str->s[str->l++] = '?'; str->s[str->l++] = ' '; str->s[str->l++] = '-'; str->s[str->l++] = ' ';
             }
-            else 
-            { 
+            else
+            {
                 str->s[str->l++] = '?'; str->s[str->l++] = ' '; str->s[str->l++] = '?'; str->s[str->l++] = ' ';
             }
         }
@@ -1116,14 +1127,21 @@ static void process_rsid_hex(convert_t *convert, bcf1_t *line, fmt_t *fmt, int i
 
 static void process_variantkey_hex(convert_t *convert, bcf1_t *line, fmt_t *fmt, int isample, kstring_t *str)
 {
+    const char *alt = NULL;
+    size_t sizealt = 0;
+    if ( line->n_allele>1 )
+    {
+        alt = line->d.allele[1];
+        sizealt = strlen(line->d.allele[1]);
+    }
     uint64_t vk = variantkey(
         convert->header->id[BCF_DT_CTG][line->rid].key,
         strlen(convert->header->id[BCF_DT_CTG][line->rid].key),
         line->pos,
         line->d.allele[0],
         strlen(line->d.allele[0]),
-        line->d.allele[1],
-        strlen(line->d.allele[1]));
+        alt,
+        sizealt);
     ksprintf(str, "%016" PRIx64 "", vk);
 }
 
@@ -1175,30 +1193,29 @@ static void process_pbinom(convert_t *convert, bcf1_t *line, fmt_t *fmt, int isa
         int al = bcf_gt_allele(gt[i]);
         if ( al > line->n_allele || al >= fmt->fmt->n ) goto invalid;
 
-        #define BRANCH(type_t, missing, vector_end) { \
-            type_t val = ((type_t *) fmt->fmt->p)[al + isample*fmt->fmt->n]; \
+        #define BRANCH(type_t, convert, missing, vector_end) { \
+            type_t val = convert(&fmt->fmt->p[(al + isample*fmt->fmt->n)*sizeof(type_t)]); \
             if ( val==missing || val==vector_end ) goto invalid; \
             else n[i] = val; \
         }
         switch (fmt->fmt->type)
         {
-            case BCF_BT_INT8:  BRANCH(int8_t,  bcf_int8_missing,  bcf_int8_vector_end); break;
-            case BCF_BT_INT16: BRANCH(int16_t, bcf_int16_missing, bcf_int16_vector_end); break;
-            case BCF_BT_INT32: BRANCH(int32_t, bcf_int32_missing, bcf_int32_vector_end); break;
+            case BCF_BT_INT8:  BRANCH(int8_t,  le_to_i8,  bcf_int8_missing,  bcf_int8_vector_end); break;
+            case BCF_BT_INT16: BRANCH(int16_t, le_to_i16, bcf_int16_missing, bcf_int16_vector_end); break;
+            case BCF_BT_INT32: BRANCH(int32_t, le_to_i32, bcf_int32_missing, bcf_int32_vector_end); break;
             default: goto invalid; break;
         }
         #undef BRANCH
     }
 
     if ( n[0]==n[1] ) kputc(n[0]==0 ? '.':'0', str);
-    else 
+    else
     {
-        double pval = n[0] < n[1] ? kf_betai(n[1], n[0] + 1, 0.5) : kf_betai(n[0], n[1] + 1, 0.5);
-        pval *= 2;
-        assert( pval-1 < 1e-10 );
-        if ( pval>=1 ) pval = 0;     // this can happen, machine precision error, eg. kf_betai(1,0,0.5)
-        else
-            pval = -4.34294481903*log(pval);
+        double pval = calc_binom_two_sided(n[0],n[1],0.5);
+
+        // convrt to phred
+        if ( pval>=1 ) pval = 0;
+        else pval = -4.34294481903*log(pval);
         kputd(pval, str);
     }
     return;
@@ -1207,7 +1224,35 @@ invalid:
     kputc('.', str);
 }
 
-static fmt_t *register_tag(convert_t *convert, int type, char *key, int is_gtf)
+static void _used_tags_add(convert_t *convert, int type, char *key)
+{
+    kstring_t str = {0,0,0};
+    ksprintf(&str,"%s/%s",type==T_INFO?"INFO":"FORMAT",key);
+    khash_str2int_inc(convert->used_tags_hash,str.s);
+    convert->nused_tags++;
+    convert->used_tags_list = (char**)realloc(convert->used_tags_list,sizeof(*convert->used_tags_list)*convert->nused_tags);
+    convert->used_tags_list[convert->nused_tags-1] = str.s;
+}
+
+
+#define _SET_NON_FORMAT_TAGS(function,key,...) \
+    if ( !strcmp("CHROM",key) ) { function(__VA_ARGS__, T_CHROM); } \
+    else if ( !strcmp("POS",key) ) { function(__VA_ARGS__, T_POS); } \
+    else if ( !strcmp("POS0",key) ) { function(__VA_ARGS__, T_POS0); } \
+    else if ( !strcmp("END",key) ) { function(__VA_ARGS__, T_END); } \
+    else if ( !strcmp("END0",key) ) { function(__VA_ARGS__, T_END0); } \
+    else if ( !strcmp("ID",key) ) { function(__VA_ARGS__, T_ID); } \
+    else if ( !strcmp("REF",key) ) { function(__VA_ARGS__, T_REF); } \
+    else if ( !strcmp("FIRST_ALT",key) ) { function(__VA_ARGS__, T_FIRST_ALT); } \
+    else if ( !strcmp("QUAL",key) ) { function(__VA_ARGS__, T_QUAL); } \
+    else if ( !strcmp("TYPE",key) ) { function(__VA_ARGS__, T_TYPE); } \
+    else if ( !strcmp("FILTER",key) ) { function(__VA_ARGS__, T_FILTER); } \
+    else if ( !strcmp("IS_TS",key) ) { function(__VA_ARGS__, T_IS_TS); } \
+    else if ( !strcmp("MASK",key) ) { function(__VA_ARGS__, T_MASK); } \
+    else if ( !strcmp("LINE",key) ) { function(__VA_ARGS__, T_LINE); }
+
+static void set_type(fmt_t *fmt, int type) { fmt->type = type; }
+static fmt_t *register_tag(convert_t *convert, char *key, int is_gtf, int type)
 {
     convert->nfmt++;
     if ( convert->nfmt > convert->mfmt )
@@ -1229,26 +1274,22 @@ static fmt_t *register_tag(convert_t *convert, int type, char *key, int is_gtf)
         int id = bcf_hdr_id2int(convert->header, BCF_DT_ID, key);
         if ( fmt->type==T_FORMAT && !bcf_hdr_idinfo_exists(convert->header,BCF_HL_FMT,id) )
         {
-            if ( !strcmp("CHROM",key) ) { fmt->type = T_CHROM; }
-            else if ( !strcmp("POS",key) ) { fmt->type = T_POS; }
-            else if ( !strcmp("POS0",key) ) { fmt->type = T_POS0; }
-            else if ( !strcmp("END",key) ) { fmt->type = T_END; }
-            else if ( !strcmp("END0",key) ) { fmt->type = T_END0; }
-            else if ( !strcmp("ID",key) ) { fmt->type = T_ID; }
-            else if ( !strcmp("REF",key) ) { fmt->type = T_REF; }
-            else if ( !strcmp("ALT",key) ) { fmt->type = T_ALT; }
-            else if ( !strcmp("FIRST_ALT",key) ) { fmt->type = T_FIRST_ALT; }
-            else if ( !strcmp("QUAL",key) ) { fmt->type = T_QUAL; }
-            else if ( !strcmp("FILTER",key) ) { fmt->type = T_FILTER; }
-            else if ( !strcmp("_CHROM_POS_ID",key) ) { fmt->type = T_CHROM_POS_ID; }
+            _SET_NON_FORMAT_TAGS(set_type,key,fmt)
+           else if ( !strcmp("ALT",key) ) { fmt->type = T_ALT; }
+           else if ( !strcmp("_CHROM_POS_ID",key) ) { fmt->type = T_CHROM_POS_ID; }
             else if ( !strcmp("RSX",key) ) { fmt->type = T_RSX; }
             else if ( !strcmp("VKX",key) ) { fmt->type = T_VKX; }
-            else if ( id>=0 && bcf_hdr_idinfo_exists(convert->header,BCF_HL_INFO,id) ) { fmt->type = T_INFO; }
+            else if ( id>=0 && bcf_hdr_idinfo_exists(convert->header,BCF_HL_INFO,id) )
+            {
+                fmt->type = T_INFO;
+                _used_tags_add(convert,T_INFO,key);
+            }
         }
         else if ( fmt->type==T_PBINOM )
         {
             fmt->id = bcf_hdr_id2int(convert->header, BCF_DT_ID, fmt->key);
             if ( !bcf_hdr_idinfo_exists(convert->header,BCF_HL_FMT, fmt->id)  ) error("No such FORMAT tag defined in the header: %s\n", fmt->key);
+            _used_tags_add(convert,T_FORMAT,key);
         }
         else if ( fmt->type==T_NPASS )
         {
@@ -1328,20 +1369,20 @@ static char *parse_tag(convert_t *convert, char *p, int is_gtf)
     kputsn(p, q-p, &str);
     if ( is_gtf )
     {
-        if ( !strcmp(str.s, "SAMPLE") ) register_tag(convert, T_SAMPLE, "SAMPLE", is_gtf);
-        else if ( !strcmp(str.s, "GT") ) register_tag(convert, T_GT, "GT", is_gtf);
-        else if ( !strcmp(str.s, "TGT") ) register_tag(convert, T_TGT, "GT", is_gtf);
-        else if ( !strcmp(str.s, "TBCSQ") ) 
+        if ( !strcmp(str.s, "SAMPLE") ) register_tag(convert, "SAMPLE", is_gtf, T_SAMPLE);
+        else if ( !strcmp(str.s, "GT") ) register_tag(convert, "GT", is_gtf, T_GT);
+        else if ( !strcmp(str.s, "TGT") ) register_tag(convert, "GT", is_gtf, T_TGT);
+        else if ( !strcmp(str.s, "TBCSQ") )
         {
-            fmt_t *fmt = register_tag(convert, T_TBCSQ, "BCSQ", is_gtf);
+            fmt_t *fmt = register_tag(convert, "BCSQ", is_gtf, T_TBCSQ);
             fmt->subscript = parse_subscript(&q);
             if ( fmt->subscript==-1 )
-            { 
+            {
                 if ( !strncmp(q,"{*}",3) ) { fmt->subscript = 0; q += 3; }
             }
             else fmt->subscript++;
         }
-        else if ( !strcmp(str.s, "IUPACGT") ) register_tag(convert, T_IUPAC_GT, "GT", is_gtf);
+        else if ( !strcmp(str.s, "IUPACGT") ) register_tag(convert, "GT", is_gtf, T_IUPAC_GT);
         else if ( !strcmp(str.s, "INFO") )
         {
             if ( *q!='/' )
@@ -1357,8 +1398,9 @@ static char *parse_tag(convert_t *convert, char *p, int is_gtf)
             while ( *q && (isalnum(*q) || *q=='_' || *q=='.') ) q++;
             if ( q-p==0 ) error("Could not parse format string: %s\n", convert->format_str);
             kputsn(p, q-p, &str);
-            fmt_t *fmt = register_tag(convert, T_INFO, str.s, is_gtf);
+            fmt_t *fmt = register_tag(convert, str.s, is_gtf, T_INFO);
             fmt->subscript = parse_subscript(&q);
+            _used_tags_add(convert,T_INFO,str.s);
         }
         else if ( !strcmp(str.s,"PBINOM") )
         {
@@ -1368,47 +1410,33 @@ static char *parse_tag(convert_t *convert, char *p, int is_gtf)
             while ( *q && *q!=')' ) q++;
             if ( q-p==0 ) error("Could not parse format string: %s\n", convert->format_str);
             kputsn(p, q-p, &str);
-            register_tag(convert, T_PBINOM, str.s, is_gtf);
+            register_tag(convert, str.s, is_gtf, T_PBINOM);
             q++;
         }
         else if ( !strcmp(str.s,"N_PASS") )
             error("N_PASS() must be placed outside the square brackets\n");
         else
         {
-            fmt_t *fmt = register_tag(convert, T_FORMAT, str.s, is_gtf);
+            fmt_t *fmt = register_tag(convert, str.s, is_gtf, T_FORMAT);
             fmt->subscript = parse_subscript(&q);
         }
     }
     else
     {
-        if ( !strcmp(str.s, "CHROM") ) register_tag(convert, T_CHROM, str.s, is_gtf);
-        else if ( !strcmp(str.s, "POS") ) register_tag(convert, T_POS, str.s, is_gtf);
-        else if ( !strcmp(str.s, "POS0") ) register_tag(convert, T_POS0, str.s, is_gtf);
-        else if ( !strcmp(str.s, "END") ) register_tag(convert, T_END, str.s, is_gtf);
-        else if ( !strcmp(str.s, "END0") ) register_tag(convert, T_END0, str.s, is_gtf);
-        else if ( !strcmp(str.s, "ID") ) register_tag(convert, T_ID, str.s, is_gtf);
-        else if ( !strcmp(str.s, "REF") ) register_tag(convert, T_REF, str.s, is_gtf);
-        else if ( !strcmp(str.s, "ALT") ) 
+        _SET_NON_FORMAT_TAGS(register_tag, str.s, convert, str.s, is_gtf)
+        else if ( !strcmp(str.s, "ALT") )
         {
-            fmt_t *fmt = register_tag(convert, T_ALT, str.s, is_gtf);
+            fmt_t *fmt = register_tag(convert, str.s, is_gtf, T_ALT);
             fmt->subscript = parse_subscript(&q);
         }
-        else if ( !strcmp(str.s, "FIRST_ALT") ) register_tag(convert, T_FIRST_ALT, str.s, is_gtf);
-        else if ( !strcmp(str.s, "QUAL") ) register_tag(convert, T_QUAL, str.s, is_gtf);
-        else if ( !strcmp(str.s, "FILTER") ) register_tag(convert, T_FILTER, str.s, is_gtf);
-        else if ( !strcmp(str.s, "QUAL") ) register_tag(convert, T_QUAL, str.s, is_gtf);
-        else if ( !strcmp(str.s, "IS_TS") ) register_tag(convert, T_IS_TS, str.s, is_gtf);
-        else if ( !strcmp(str.s, "TYPE") ) register_tag(convert, T_TYPE, str.s, is_gtf);
-        else if ( !strcmp(str.s, "MASK") ) register_tag(convert, T_MASK, str.s, is_gtf);
-        else if ( !strcmp(str.s, "LINE") ) register_tag(convert, T_LINE, str.s, is_gtf);
-        else if ( !strcmp(str.s, "_CHROM_POS_ID") ) register_tag(convert, T_CHROM_POS_ID, str.s, is_gtf);
-        else if ( !strcmp(str.s, "_GT_TO_PROB3") ) register_tag(convert, T_GT_TO_PROB3, str.s, is_gtf);
-        else if ( !strcmp(str.s, "_PL_TO_PROB3") ) register_tag(convert, T_PL_TO_PROB3, str.s, is_gtf);
-        else if ( !strcmp(str.s, "_GP_TO_PROB3") ) register_tag(convert, T_GP_TO_PROB3, str.s, is_gtf);
-        else if ( !strcmp(str.s, "_GT_TO_HAP") ) register_tag(convert, T_GT_TO_HAP, str.s, is_gtf);
-        else if ( !strcmp(str.s, "_GT_TO_HAP2") ) register_tag(convert, T_GT_TO_HAP2, str.s, is_gtf);
-        else if ( !strcmp(str.s, "RSX") ) register_tag(convert, T_RSX, str.s, is_gtf);
-        else if ( !strcmp(str.s, "VKX") ) register_tag(convert, T_VKX, str.s, is_gtf);
+        else if ( !strcmp(str.s, "_CHROM_POS_ID") ) register_tag(convert, str.s, is_gtf, T_CHROM_POS_ID);
+        else if ( !strcmp(str.s, "_GT_TO_PROB3") ) register_tag(convert, str.s, is_gtf, T_GT_TO_PROB3);
+        else if ( !strcmp(str.s, "_PL_TO_PROB3") ) register_tag(convert, str.s, is_gtf, T_PL_TO_PROB3);
+        else if ( !strcmp(str.s, "_GP_TO_PROB3") ) register_tag(convert, str.s, is_gtf, T_GP_TO_PROB3);
+        else if ( !strcmp(str.s, "_GT_TO_HAP") ) register_tag(convert, str.s, is_gtf, T_GT_TO_HAP);
+        else if ( !strcmp(str.s, "_GT_TO_HAP2") ) register_tag(convert, str.s, is_gtf, T_GT_TO_HAP2);
+        else if ( !strcmp(str.s, "RSX") ) register_tag(convert, str.s, is_gtf, T_RSX);
+        else if ( !strcmp(str.s, "VKX") ) register_tag(convert, str.s, is_gtf, T_VKX);
         else if ( !strcmp(str.s,"PBINOM") ) error("Error: PBINOM() is currently supported only with FORMAT tags. (todo)\n");
         else if ( !strcmp(str.s, "INFO") )
         {
@@ -1419,14 +1447,15 @@ static char *parse_tag(convert_t *convert, char *p, int is_gtf)
                 while ( *q && (isalnum(*q) || *q=='_' || *q=='.') ) q++;
                 if ( q-p==0 ) error("Could not parse format string: %s\n", convert->format_str);
                 kputsn(p, q-p, &str);
-                fmt_t *fmt = register_tag(convert, T_INFO, str.s, is_gtf);
+                fmt_t *fmt = register_tag(convert, str.s, is_gtf, T_INFO);
                 fmt->subscript = parse_subscript(&q);
+                _used_tags_add(convert,T_INFO,str.s);
             }
             else
-                register_tag(convert, T_INFO, NULL, is_gtf);    // the whole INFO
+                register_tag(convert, NULL, is_gtf, T_INFO);    // the whole INFO
         }
         else if ( !strcmp(str.s, "FORMAT") )
-             register_tag(convert, T_FORMAT, NULL, 0);
+             register_tag(convert, NULL, 0, T_FORMAT);
         else if ( !strcmp(str.s,"N_PASS") )
         {
             if ( *q!='(' ) error("Could not parse the expression: %s\n", convert->format_str);
@@ -1441,12 +1470,13 @@ static char *parse_tag(convert_t *convert, char *p, int is_gtf)
             }
             if ( q-p==0 || nopen ) error("Could not parse format string: %s\n", convert->format_str);
             kputsn(p, q-p-1, &str);
-            register_tag(convert, T_NPASS, str.s, is_gtf);
+            register_tag(convert, str.s, is_gtf, T_NPASS);
         }
         else
         {
-            fmt_t *fmt = register_tag(convert, T_INFO, str.s, is_gtf);
+            fmt_t *fmt = register_tag(convert, str.s, is_gtf, T_INFO);
             fmt->subscript = parse_subscript(&q);
+            _used_tags_add(convert,T_INFO,str.s);
         }
     }
     free(str.s);
@@ -1470,7 +1500,7 @@ static char *parse_sep(convert_t *convert, char *p, int is_gtf)
         q++;
     }
     if ( !str.l ) error("Could not parse format string: %s\n", convert->format_str);
-    register_tag(convert, T_SEP, str.s, is_gtf);
+    register_tag(convert, str.s, is_gtf, T_SEP);
     free(str.s);
     return q;
 }
@@ -1481,6 +1511,7 @@ convert_t *convert_init(bcf_hdr_t *hdr, int *samples, int nsamples, const char *
     convert->header = hdr;
     convert->format_str = strdup(format_str);
     convert->max_unpack = BCF_UN_STR;
+    convert->used_tags_hash = khash_str2int_init();
 
     int i, is_gtf = 0;
     char *p = convert->format_str;
@@ -1490,7 +1521,7 @@ convert_t *convert_init(bcf_hdr_t *hdr, int *samples, int nsamples, const char *
         switch (*p)
         {
             case '[': is_gtf = 1; p++; break;
-            case ']': is_gtf = 0; register_tag(convert, T_SEP, NULL, 0); p++; break;
+            case ']': is_gtf = 0; register_tag(convert, NULL, 0, T_SEP); p++; break;
             case '%': p = parse_tag(convert, p, is_gtf); break;
             default:  p = parse_sep(convert, p, is_gtf); break;
         }
@@ -1521,6 +1552,13 @@ void convert_destroy(convert_t *convert)
         if ( convert->fmt[i].destroy ) convert->fmt[i].destroy(convert->fmt[i].usr);
         free(convert->fmt[i].key);
     }
+    if ( convert->nused_tags )
+    {
+        for (i=0; i<convert->nused_tags; i++) free(convert->used_tags_list[i]);
+        free(convert->used_tags_list);
+    }
+    khash_str2int_destroy(convert->used_tags_hash);
+    free(convert->print_filtered);
     free(convert->fmt);
     free(convert->undef_info_tag);
     free(convert->dat);
@@ -1535,13 +1573,19 @@ int convert_header(convert_t *convert, kstring_t *str)
     int i, icol = 0, l_ori = str->l;
     bcf_hdr_t *hdr = convert->header;
 
-    // Supress the header output if LINE is present
+    // Suppress the header output if LINE is present
     for (i=0; i<convert->nfmt; i++)
         if ( convert->fmt[i].type == T_LINE ) break;
     if ( i!=convert->nfmt )
         return str->l - l_ori;
 
-    kputs("# ", str);
+    // Header formatting becomes problematic when the formatting expression contains a newline.
+    // Simple cases like
+    //      -f'[%CHROM %POS %SAMPLE\n]'
+    // can be handled quite easily with has_fmt_newline. Note this will not work if multiple newlines
+    // are present.
+    int has_fmt_newline = 0;
+    kputc('#', str);
     for (i=0; i<convert->nfmt; i++)
     {
         // Genotype fields
@@ -1556,12 +1600,39 @@ int convert_header(convert_t *convert, kstring_t *str)
                 {
                     if ( convert->fmt[k].type == T_SEP )
                     {
-                        if ( convert->fmt[k].key ) kputs(convert->fmt[k].key, str);
+                        if ( convert->fmt[k].key )
+                        {
+                            char *tmp = convert->fmt[k].key;
+                            while ( *tmp )
+                            {
+                                if ( *tmp=='\n' ) has_fmt_newline = 1;
+                                else kputc(*tmp,str);
+                                tmp++;
+                            }
+                        }
+                    }
+                    else if ( convert->header_samples )
+                    {
+                        icol++;
+                        if ( !convert->no_hdr_indices ) ksprintf(str,"[%d]",icol);
+                        ksprintf(str,"%s:%s", hdr->samples[ks], convert->fmt[k].key);
                     }
-                    else if ( convert->fmt[k].type == T_SAMPLE )
-                        ksprintf(str, "[%d]%s", ++icol, convert->fmt[k].key);
                     else
-                        ksprintf(str, "[%d]%s:%s", ++icol, hdr->samples[ks], convert->fmt[k].key);
+                    {
+                        icol++;
+                        if ( !convert->no_hdr_indices ) ksprintf(str,"[%d]",icol);
+                        ksprintf(str,"%s", convert->fmt[k].key);
+                    }
+                }
+                if ( has_fmt_newline )
+                {
+                    if ( !convert->header_samples ) break;
+
+                    // this is unfortunate: the formatting expression breaks the per-sample output into separate lines,
+                    // therefore including a sample name in the header makes no sense anymore
+                    convert->header_samples = 0;
+                    str->l = l_ori;
+                    return convert_header(convert, str);
                 }
             }
             i = j-1;
@@ -1573,8 +1644,11 @@ int convert_header(convert_t *convert, kstring_t *str)
             if ( convert->fmt[i].key ) kputs(convert->fmt[i].key, str);
             continue;
         }
-        ksprintf(str, "[%d]%s", ++icol, convert->fmt[i].key);
+        icol++;
+        if ( !convert->no_hdr_indices ) ksprintf(str,"[%d]",icol);
+        ksprintf(str,"%s", convert->fmt[i].key);
     }
+    if ( has_fmt_newline ) kputc('\n',str);
     return str->l - l_ori;
 }
 
@@ -1598,7 +1672,7 @@ int convert_line(convert_t *convert, bcf1_t *line, kstring_t *str)
     str->l = 0;
     for (i=0; i<convert->nfmt; i++)
     {
-        // Genotype fields. 
+        // Genotype fields.
         if ( convert->fmt[i].is_gt_field )
         {
             int j = i, js, k;
@@ -1611,7 +1685,17 @@ int convert_line(convert_t *convert, bcf1_t *line, kstring_t *str)
             {
                 // Skip samples when filtering was requested
                 int ks = convert->samples[js];
-                if ( convert->subset_samples && *convert->subset_samples && !(*convert->subset_samples)[ks] ) continue;
+                if ( convert->subset_samples && *convert->subset_samples && !(*convert->subset_samples)[ks] )
+                {
+                    if ( !convert->print_filtered ) continue;
+
+                    for (k=i; k<j; k++)
+                        if ( convert->fmt[k].type==T_SEP )
+                            convert->fmt[k].handler(convert, line, &convert->fmt[k], ks, str);
+                        else
+                            kputs(convert->print_filtered, str);
+                    continue;
+                }
 
                 // Here comes a hack designed for TBCSQ. When running on large files,
                 // such as 1000GP, there are too many empty fields in the output and
@@ -1619,7 +1703,7 @@ int convert_line(convert_t *convert, bcf1_t *line, kstring_t *str)
                 // anything to the string, we trim all genotype fields enclosed in square
                 // brackets here. This may be changed in future, time will show...
                 size_t l_start = str->l;
-            
+
                 for (k=i; k<j; k++)
                 {
                     if ( convert->fmt[k].type == T_MASK )
@@ -1651,13 +1735,43 @@ int convert_line(convert_t *convert, bcf1_t *line, kstring_t *str)
     return str->l - l_ori;
 }
 
+static void force_newline_(convert_t *convert)
+{
+    int i, has_newline = 0;
+    for (i=0; i<convert->nfmt; i++)
+    {
+        if ( !convert->fmt[i].key ) continue;
+        char *tmp = convert->fmt[i].key;
+        while (*tmp)
+        {
+            if ( *tmp=='\n' ) { has_newline = 1; break; }
+            tmp++;
+        }
+        if ( has_newline ) break;
+    }
+    if ( has_newline ) return;
+
+    // A newline is not present, force it. But where to add it? Always at the end.
+    //
+    // Briefly, in 1.18, we considered the following automatic behavior, which for
+    // per-site output it would add it at the end of the expression and for per-sample
+    // output it would add it inside the square brackets:
+    //           -f'%CHROM[ %SAMPLE]\n'
+    //           -f'[%CHROM %SAMPLE\n]'
+    //
+    // However, this is an annoyance for users, as it is not entirely clear what
+    // will happen unless one understands the internals well (#1969)
+
+    register_tag(convert, "\n", 0, T_SEP);
+}
+
 int convert_set_option(convert_t *convert, enum convert_option opt, ...)
 {
     int ret = 0;
     va_list args;
 
     va_start(args, opt);
-    switch (opt) 
+    switch (opt)
     {
         case allow_undef_tags:
             convert->allow_undef_tags = va_arg(args, int);
@@ -1665,6 +1779,19 @@ int convert_set_option(convert_t *convert, enum convert_option opt, ...)
         case subset_samples:
             convert->subset_samples = va_arg(args, uint8_t**);
             break;
+        case header_samples:
+            convert->header_samples = va_arg(args, int);
+            break;
+        case print_filtered:
+            convert->print_filtered = strdup(va_arg(args, char*));
+            break;
+        case force_newline:
+            convert->force_newline = va_arg(args, int);
+            if ( convert->force_newline ) force_newline_(convert);
+            break;
+        case no_hdr_indices:
+            convert->no_hdr_indices = va_arg(args, int);
+            break;
         default:
             ret = -1;
     }
@@ -1677,3 +1804,13 @@ int convert_max_unpack(convert_t *convert)
     return convert->max_unpack;
 }
 
+int convert_is_tag_used(convert_t *convert, char *tag)
+{
+    return khash_str2int_has_key(convert->used_tags_hash, tag);
+}
+const char **convert_list_used_tags(convert_t *convert, int *ntags)
+{
+    *ntags = convert->nused_tags;
+    return (const char **)convert->used_tags_list;
+}
+
